; *************************************
; ZX80 3K Space Invaders - Main Program
; *************************************
; (c)1981 Macronics
;
; RAM required: 3K
;
; Disassembled by Paul Farrow, www.fruitcake.plus.com
; Initial version 17th December 2009 (revised 16th April 2011)


; =====
; Notes
; =====

; The program spans from $4000 (16384) to $42C6 (17094) inclusive.
;
; The Display file is located at $44FF (17663). It begins with a newline character ($76), followed by 24 rows of
; 31 columns. Each column is terminated with a newline character ($76).
;
; The following program variables are used:
;
;               $44E0    = Top of stack. Holds a sequence counter which indicates what action to perform at each frame.
;                          The bit combinations have the following meanings:
;                            00 = Move a row of invaders.
;                            01 = Move the invader bomb and laser base missile.
;                            10 = Read keyboard and move laser base.
;                            11 = Check whether to drop a new invader bomb.
; IY+$19/1A     $44F4/F5 = Invader bomb display file position or $00xx for none.
; IY+$1B/1C     $44F6/F7 = Laser base missile display file position or $00xx for none.
; IY+$1D        $44F8    = Holds the number of invaders currently on screen (up to a maximum of 16).
; IY+$1E        $44F9    = Not used.
; IY+$20        $44FB    = A flag used to indicate whether an invader has been selected to drop a bomb ($FF=Invader selected, else $00).
; IY+$21/22     $44FC/FD = Laser base display file position.
; IY+$23        $44FE    = Holds the number of scan lines to produce. It is used by the ROM display routines, which addresses this
;                          location using IY+$23.
;               $44FF    = Display file, ending at $47FF.


; ====================
; Assembler Directives
; ====================

; For use with the TASM assembler.

#define DEFB .BYTE
#define DEFW .WORD
#define DEFM .TEXT
#define DEFS .FILL
#define EQU  .EQU
#define ORG  .ORG
#define END  .END

; ======================================================================================================================================================

        ORG     $4000

; -------------------
; Program Entry Point
; -------------------

L4000:  LD      SP,$44E0                ; Set up a stack at 17632 (+1248).

        LD      IY,$44DB                ; Base address of program variables.
        LD      (IY+$1A),$00            ; Clear the display file position of the invader bomb, i.e mark it as no active bomb.
        LD      (IY+$1C),$00            ; Clear the display file position of the laser base missile, i.e. being mark as no active missile.
        LD      (IY+$21),$B2            ; Set the default laser base position (row 21, column 18), situated under the third shield.

; Initialise the display file to all black.

        LD      HL,$4500                ; Start of the display file.
        LD      A,$80                   ; Black character.
        LD      B,$00                   ; Loop counter of 256.

L401A:  LD      (HL),A                  ; Fill 3 bytes (3*256=768=24*32).
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        DJNZ    L401A                   ; Repeat for all characters of the display file.

; Draw the four shields. Each shield is 4 characters wide by 5 high. They are located at columns 3, 10, 17 and 24.

        LD      HL,$46E3                ; Start location within display file to store the shields (row 15, column 3).

L4025:  LD      A,$00                   ; The character used to draw the shields.
        LD      B,$04                   ; There are 4 shields.

L4029:  LD      (HL),A                  ; Draw a row of one shield.
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        LD      (HL),A

        INC     HL                      ; Advance to the start of the the next shield within this row.
        INC     HL
        INC     HL
        INC     HL
        DJNZ    L4029                   ; Repeat for all 4 shields in this row.

; Advance to the next row.

        LD      A,L                     ; Has the end of the 5 row been reached?
        CP      $7F                     ; Tested by Comparing to address $477F.

        INC     HL                      ; Advance to the start of the first shield on the next row.
        INC     HL
        INC     HL
        INC     HL
        JR      NZ,L4025                ; Draw the next row if required.

; The shields have now been drawn.

        NOP                             ; Unused. Do nothing.

; Insert a newline character at the start of the display file and at the end of each row, i.e. as the 32nd character of each row.

        LD      DE,$0020                ; Offset to the next display file row.
        LD      HL,$44FF                ; Point to the byte before the display file.
        LD      B,$19                   ; 25 characters.

L4048:  LD      (HL),$76                ; Insert the row terminator.
        ADD     HL,DE                   ; Advance to the next row.
        DJNZ    L4048                   ; Repeat for all rows.

; Draw the score at the bottom right hand corner.

        LD      B,$04                   ; The score is displayed as 4 digits.
        LD      HL,$47FB                ; Row 23, column 27.

L4052:  LD      (HL),$1C                ; Insert a character '0'.
        INC     HL                      ; Next location.
        DJNZ    L4052                   ; Repeat for all digits.

; Draw the last score at the bottom left hand corner.

        LD      BC,$0004                ; The last score is displayed as 4 digits.
        LD      HL,L416C                ; Address of the store holding the last score.
        LD      DE,$47E0                ; Row 23, column 0.
        LDIR                            ; Display the last score.

; The main program loop used to commence a new wave of invaders.

L4062:  LD      HL,$46BD                ; Set the address of the next row within the invader area of the display file to start shifting from as row 13, column 29.
        LD      C,$08                   ; The number of shifts remaining before the invaders reached the edge of the display and will require moving down a row. B will hold $00.
        EXX

; Draw the space invaders.
;
; .'4'.  = Graphic $08, Character 1, Graphic $88.
;
;  '3'   = Graphic $05, Character 2, Graphic $04.
;
; '.2.'  = Graphic $88, Character 3, Graphic $08.
;
;  .1.   = Graphic $07, Character 4, Graphic $06.

; Draw the first row of space invaders.

        LD      DE,$0888                ; The left and right characters of the invader.
        LD      A,$20                   ; The middle character is '4'.
        LD      HL,$4521                ; Row 1, column 1.
        CALL    L40D0                   ; Draw a row of space invaders.

; Draw the second row of space invaders.

        LD      DE,$0504                ; The left and right characters of the invader.
        DEC     A                       ; The middle character is '3'.
        LD      L,$61                   ; Row 3, column 1.
        CALL    L40D0                   ; Draw a row of space invaders.

; Draw the third row of space invaders.

        LD      DE,$8808                ; The left and right characters of the invader.
        DEC     A                       ; The middle character is '2'.
        LD      L,$A1                   ; Row 5, column 1.
        CALL    L40D0                   ; Draw a row of space invaders.

; Draw the fourth row of space invaders.

        LD      DE,$0706                ; The left and right characters of the invader.
        DEC     A                       ; The middle character is '1'.
        LD      L,$E1                   ; Row 7, column 1.
        CALL    L40D0                   ; Draw a row of space invaders.

        LD      (IY+$1D),$10            ; Set the number of invaders on screen (initially there are 4 rows of 4 invaders, i.e. 16 in total).

; A PAL TV picture consists of 25 frames per second, consisting of 625 scan lines. Each frame is divided into 2 interlaced fields, referred to
; as odd and even fields, with each field consisting of 312.5 scan lines. A field occupies 20ms and hence each scan line takes 64us.
; To produce a steady TV picture, the display routine within the ZX80 ROM has been carefully structured so that each iteration of the loop
; takes exactly 20ms to execute. The ZX80 is driven by a 6.5MHz ceramic resonator/filter, which is halved and tuned down to approximately 3.21MHz.
; Each iteration of display loop within the ROM takes 64167 T-states and this spans 20ms. Therefore each T-states lasts 311.67ns, and hence the
; CPU clock frequency assumed by the ZX80 is 3.20835MHz (which is close to the 3.2MHz marked on the ZX80 schematic). Ceramic resonators do not
; have a tight tolerance and so some variation around this figure is to be expected. The ZX80 ROM outputs 303 scan lines (56 in the top border,
; 192 in the main area and 55 in the bottom border), with each requiring 207 T-states to generate. Therefore 303 * 207 = 62721 T-states are
; required to span these scan lines, with an additional 27 T-states used to return control to the display loop. The remaining 1419 T-states
; are used to read the keyboard and update the FRAMES system variable. Note that with a clock frequency of 3.20835MHz, each scan line takes 64.52us.
; This means that the TV field produced by the ZX80 spans 309.99 of these slightly longer scan lines.
;
; The 3K Space Invaders game contains a display loop that outputs 287 scan lines per TV field (48 in the top border, 192 in the main display area,
; and 47 in the bottom border). This frees up 16 * 207 = 3312 T-states, making a total of 4742 T-states (equating to 1.478ms) available per TV field
; for running the game. To achieve a steady display, the functionality performed at each field must be structured to fill this time as accurately
; as possible. The game code is divided into four sections, and these are cycled through one per TV field. It takes 146 T-states to run the sequence
; control loop, leaving 4596 T-states (1.433ms) for each game section. The game sections actually take the following durations:
;
; Part 1 = 4597 T-states (Range 4589 to 4607 T-states)  -  Move row of invaders.
; Part 2 = 4598 T-states (Range 4571 or 4899 T-states)  -  Move invader bomb and laser base missile.
; Part 3 = 4597 T-states (Range 4595 to 4598 T-states)  -  Read keyboard and move laser base.
; Part 4 = 4598 T-states (Range 4570 to 4598 T-states)  -  Check whether to drop new invader bomb.
;
; There is a range of values for each section due to small differences that can occur between different paths within those routines. The values
; are all close enough to the target value of 1.433ms that any difference simply causes the next field to be displayed slightly sooner or later
; than the desired 20ms rate. Although it is desirable to produce TV fields at a precise rate of 20ms, a small amount of variation is tolerable
; and television sets can cope with this. The target duration of 1.433ms is based on the CPU speed assumed by the ZX80 ROM, but this may vary
; slightly between machines due to tolerance of the resonator and so is only a guide to the time available.
;
; The 3K Space Invaders game is written to generate a 50Hz display field rate suitable for the UK. The ZX80 ROM generates a 60Hz display field rate
; by reducing the number of top and bottom border lines by 24 each. The game can be modified to support a 60Hz display rate by modifying the
; following line as indicated.

L4092:  LD      (IY+$23),$30            ; 19    There will be 48 scan lines above the ZX80 TV display area to achieve a 50Hz field rate.
                                        ;       [Change this to 24 ($18) scan lines for USA 60Hz]

        POP     HL                      ; 10    Fetch the byte at $44E1, which is used as a sequence counter to determine what action to perform this frame.
        INC     L                       ; 4     Increment the counter.
        PUSH    HL                      ; 11    Store the updated counter.

        IN      A,($FE)                 ; 11    Turn on the vertical sync generator. It be will be turned off approximately 23 scan lines from now.
                                        ;       [Any I/O read to an address with A0 low will turn on the vertical sync generator]

; The sequence counter determines the action to perform on this TV field.

        LD      A,L                     ; 4     Fetch the sequence counter.
        AND     $03                     ; 7     Keep only the lower 2 bits, forming 4 possibilities to consider.
        CALL    Z,L4110                 ; 17/10 Move a row of invaders (typical 4597 T-states, though ranges from 4589 to 4607 T-states).

        CP      $01                     ; 7
        CALL    Z,L4170                 ; 17/10 Move the invader bomb and laser base missile (typically 4598 T-states, though ranges from 4571 to 4898 T-states).

        CP      $02                     ; 7
        CALL    Z,L4209                 ; 17/10 Read the keyboard and move the laser base (typically 4597 T-states, though ranges from 4595 to 4598 T-states).

        CP      $03                     ; 7
        CALL    Z,L425D                 ; 17/10 Check whether to drop a new invader bomb (typically 4598 T-states, though ranges from 4570 to 4598 T-states).

; Generate the top TV border and main area.

        OUT     ($FE),A                 ; 11    Turn off the vertical sync generator.
                                        ;       [Note that the ZX80 ROM uses an OUT ($FF),A instruction to do this, but any I/O OUT will do]

; The first scan line produced by the program is too short but this does not cause a problem for TV sets since all subsequent scan lines are produced correctly.
; The first scan line is too short by 18 T-states. The problem can be improved by loading a value of $E9 into the A register instead of $ED. This will introduce
; an additional 16 T-states, and so reduces the error to -2 T-states. The A register hold a counter that is incremented until it reaches 0, at which point the start
; of the next scan line is triggered. Each decrement of the value loaded into the A register introduces a 4 T-states delay, and hence decrementing the value by 4
; introduces a 16 T-states delay.
;
; The ZX80 ROM takes 38 T-states from issuing OUT ($FF),A to before the CALL $01AD. The routine below takes 24 T-states, and hence is 14 T-states too short.
; In addition, the A register is loaded with $EC by the ROM compared to $ED by the routine below and this introduces a further 4 T-state deficit to make 18 T-states
; in total. If the A register is loaded with $E9 then this would introduce a 4*4=16 T-state delay, thereby reducing the deficit to 2 T-states.

        LD      A,$ED                   ; 7     Timing counter for the time left to generate for the first scan line of the top border. Thereafter set to $DD by the ROM.
        LD      B,$19                   ; 7     There are 25 HALTs within the display file, i.e. 24 rows and 1 top border.
        LD      HL,$C4FF                ; 10    Point to the start of the display file, i.e. the initial HALT representing the top border, but with bit 15 of the address set.
        CALL    $01AD                   ; 17    DISP-2. Produce the top border and main area.

; Generate the bottom TV border.

; The ZX80 ROM would take 52 T-states after returning from CALL $01AD to before the next CALL $01AD. This routine takes 40 T-states, and
; hence is 12 T-states too short. The ZX80 ROM loads the A register with $F3 and so this routine loads it with $F0 to compensate for the deficit.
; This difference of $F3-$F0=$03 introduces 3x4=12 T-states and hence eliminates the deficit.

        LD      A,$F0                   ; 7     Timing counter for the time left to generate for the first scan line of the bottom border. Thereafter set to $DD by the ROM.
        INC     B                       ; 4     Increment the row count, i.e. B holds $01.
        DEC     HL                      ; 6     Decrement the screen address, i.e. point to the last HALT.
        DEC     (IY+$23)                ; 23    Decrement the blank line counter, i.e. there will be 47 scan lines below the ZX80 TV display area.

        CALL    $01AD                   ; 17    DISP-2. Produce the bottom border.

        JR      L4092                   ; 12    Repeat the main program loop for the current wave of invaders.

; ----------------------
; All Invaders Destroyed
; ----------------------
; This routine is jumped to once all 16 invaders in the wave have been destroyed.

L40C9:  NOP                             ; 4     Perform a delay.

        LD      (HL),$80                ; 10    Ensure the missile has been blanked [Redundant].

        POP     IX                      ; 14    Drop the subroutine return address.
        JP      L4062                   ; 10    Repeat the main program loop to commence a new wave of invaders.

; --------------------------
; Draw Row of Space Invaders
; --------------------------
; This routine draws a full row of space invaders.
; Entry: HL=Display file address.
;        D=Left character of the space invader.
;        A=Middle character of the space invader.
;        E=Right character of the space invader.

L40D0:  LD      B,$04                   ; There are four invaders per row.

L40D2:  LD      (HL),D                  ; Draw the left character.
        INC     HL
        LD      (HL),A                  ; Draw the middle character.
        INC     HL
        LD      (HL),E                  ; Draw the right character.

        INC     HL                      ; Advance to the next invader position.
        INC     HL
        INC     HL
        INC     HL
        DJNZ    L40D2                   ; Repeat for all invaders on this row.

        RET

; -------------------------------------
; Laser Base Missile Has Hit An Invader
; -------------------------------------
; An invader has been hit so erase it and add its numerical value to the score.
; The routine takes a variable length of time depending on how many digits are affected by increasing the score.
; The time taken by the routine is typically 365 T-states, but can be up to 692 T-states.
; The routines tests whether all invaders in the wave have been destroyed.
; The invaders are made up of the following character codes:
;
; $08 $20 $88   or   $05 $1F $04   or   $88 $1E $08   or   $07 $1D $06
;
; Entry: HL=Address within the display file where the laser base missile hit the invader.

L40DE:  LD      B,$04                   ; 7     Four character positions are considered for the invader.
                                        ;       [This includes the position to the right of the invader].

        DEC     L                       ; 4     Point to 2 cells before the collision position. If the invader was hit on its right
        DEC     L                       ; 4     most character then HL will now be pointing at its left most character.

; Enter a loop to erase the invader and to add its 'number' to the score.

L40E2:  LD      A,(HL)                  ; 7     Fetch the display file location.
        SUB     $1C                     ; 7     Has the central character been found, i.e. it is a numeric digit?
        JP      M,L4105                 ; 10    Jump ahead if not.

        CP      $25                     ; 7     Has the central character been found, i.e. it is a numeric digit?
        JP      P,L4105                 ; 10    Jump ahead if not.

; The central character of the invader was hit, so increment the score based on the number forming the middle character of the invader.

        LD      DE,$47FE                ; 10    Point to the last digit of the score, at row 23, column 30.
        LD      C,A                     ; 4     Note the numeric value of the invader.

L40F1:  LD      A,(DE)                  ; 7     Fetch the score digit.
        ADD     A,C                     ; 4     Add the number forming the middle character of the invader.
        CP      $26                     ; 7     Has '9' been surpassed?
        JP      M,L4104                 ; 10    Jump ahead if not.

; A carry has occurred in this digit of the score.

        SUB     $0A                     ; 7     Reduce the digit by 10.
        LD      (DE),A                  ; 7     Update the score digit within the display file.

        LD      C,$01                   ; 7     A value of 1 is to be added to the previous score digit.
        DEC     E                       ; 4     Point to the previous score digit within the display file.
        LD      A,E                     ; 4
        CP      $FA                     ; 7     Pointing before the first score digit?
        JR      NZ,L40F1                ; 12/7  Jump back if not to process the previous score digit.

; The addition of the invader value has caused an overflow from 9999. The A register holds a value of $FA
; at this point and the instruction at $4104 will set this into the display file causing corruption of the
; display file. It would have been better to perform a JR $4105 and hence leave the score wrapped back at 0000.

        INC     E                       ; 4     Point at the first score digit.

L4104:  LD      (DE),A                  ; 7     Update the score digit within the display file.

L4105:  LD      (HL),$80                ; 10    Blank this character of the invader.
        INC     L                       ; 4     Point to the next character of the invader.
        DJNZ    L40E2                   ; 13/8  Repeat for the next character of the invader.

        DEC     (IY+$1D)                ; 23    Decrement the count of the number of invaders left on screen.
        JR      Z,L40C9                 ; 12/7  Jump if all invaders have been destroyed to commence a new wave of invaders.

        RET                             ; 10

; -------------
; Move Invaders
; -------------
; This routine moves all the invaders, handling the toggling of side characters for invader rows '4' and '2'.
; At each invocation, it shifts all characters on a particular row by one position left or right. All characters
; on a row are shifted each time, whether blank or not, and hence this ensures that the routine completes after a
; fixed duration of time. The area that the invaders can occupy spans rows 0 to 14, although once the invaders
; move onto row 14 the game is over. On the odd rows the invaders are shifted to the right, and on the even rows
; they are shifted to the left. Rows are shifted in decending order, beginning with row 13 and ending with row 0.
; Initially the invaders are located at the left hand side of the screen and so are shifted to the right. The
; invaders are shifted by eight columns before moving down a row. The invaders cannot occupy columns 0 or 31, and
; so these always filled with black space. This black space is shifted along with the invaders, thereby overwriting
; the 'trail' character left by an invader. The black space is not overwritten itself.
; Entry: HL'=The address of the next row within the invader area of the display file to start shifting from.
;        C'=The number of shifts remaining before the invaders reached the edge of the display and will require moving down.
;        B'=$00.
; The routine takes between 4589 T-states and 4607 T-states, though the typical time is 4597 T-states.

L4110:  EXX                             ; 4     Fetch invader area variables.

        LD      A,L                     ; 4     Point to the last display file address of the invaders.
        AND     $20                     ; 7     Extract the row address. Is it odd or even?
        JR      Z,L411B                 ; 12/7  Jump if even to move invaders to the left.

; Move the invaders to the right.

        LD      DE,$FFFF                ; 10    A value of -1 will be subtracted, thereby moving the invaders to the right.
        JR      L411F                   ; 12    Jump ahead to continue.

; Move the invaders to the left.

L411B:  LD      DE,$0001                ; 10    A value of 1 will be subtracted, thereby moving the invaders to the left.

        LD      A,(HL)                  ; 7     Perform a delay.

L411F:  LD      A,(HL)                  ; 7     Fetch a character from the invader area.

; Enter a loop to shift all characters on the current row.

L4120:  CP      $08                     ; 7     Does the location contain a diagonal-to-right side character of an invader?
        JR      NZ,L4129                ; 12/7  Jump ahead if not.

; A diagonal side character of an invader has been found so toggle it.
; Following this path results in a quicker duration by 4 T-states.

        LD      A,(HL)                  ; 7     Perform a delay.
        ADD     A,$80                   ; 7     Toggle the character.
        JR      L4130                   ; 12    Jump ahead to continue.

L4129:  CP      $88                     ; 7     Does the location contain a diagonal-to-left side character of an invader?
        NOP                             ; 4     Perform a delay.
        JR      NZ,L4130                ; 12/7  Jump ahead if not. Following this jump results in a quicker path by 2 T-states.

; A diagonal side character of an invader has been found so toggle it.

        ADD     A,$80                   ; 7     Toggle the character.

L4130:  INC     C                       ; 4     Is it time to move the invaders down to the next row?
        DEC     C                       ; 4
        JR      Z,L413C                 ; 12/7  Jump ahead if so.

; It is not yet time to move the invaders down, so shift a character on the current row over.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Add the offset to generate the address to move the character to.
        LD      (HL),A                  ; 7     Shift the character.

        ADD     HL,DE                   ; 11    Point back to the original address.
        NOP                             ; 4     Perform a delay.
        JR      L4144                   ; 12    Jump ahead to continue.

; It is time to move the invaders down.

L413C:  SET     5,C                     ; 8     This sets C to hold 32, i.e. the offset to the row below.
        ADD     HL,BC                   ; 11    Generate the address of the next row.
        LD      (HL),A                  ; 7     Shift the invader character.

        SBC     HL,BC                   ; 15    Point back to the original row.

        LD      C,$00                   ; 7     Perform a delay.

; A character on the row has been shifted. Test whether all characters in the current row have been shifted.

L4144:  ADD     HL,DE                   ; 11    Point to the next location to shift.
        LD      A,(HL)                  ; 7
        CP      $76                     ; 7     Has the end/start of the row been reached?
        JR      NZ,L4120                ; 12/7  Jump back if not to move the next character on this row.

; All characters on the current row have been shifted.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point back to the last character shifted within the row.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point to the edge position within the row.

        LD      DE,$FFE0                ; 10    Offset of -32.
        ADD     HL,DE                   ; 11    Point to the row above.

        LD      A,H                     ; 4     Has the top of the display file been exceeded?
        CP      $44                     ; 7
        JR      NZ,L4165                ; 12/7  Jump ahead if not, i.e. there are further rows to be shifted next time.

; The whole of the invader area has been shifted, i.e. one cycle has been completed.

        LD      HL,$46BD                ; 10    Reset the next display file position to shift back to row 13, column 29.
        INC     C                       ; 4
        DEC     C                       ; 4     Have the invaders been moved to the edge of the display?
        JR      NZ,L4162                ; 12/7  Jump ahead if not. Following this jump results in a quicker path by 2 T-states.

; The invaders have been shifted enough times such that they are now at the edge of the display.

        LD      C,$09                   ; 7     Reset the number of shifts, including one to cancel out the immediate decrement below.

L4162:  DEC     C                       ; 4     Decrement the count of the number of columns the invaders have been shifted.
        JR      L4169                   ; 12    Jump ahead to continue.

; A complete row of the invader area has been shifted.

L4165:  LD      B,$03                   ; 7     Perform a delay to match the other paths through this routine.

L4167:  DJNZ    L4167                   ; 13/8

L4169:  EXX                             ; 4     Save the invader area variables.

        NOP                             ; 4     Perform a delay.
        RET                             ; 10

; ----------------
; Last Score Store
; ----------------
; Holds the digits of the last score.

L416C:  DEFB    $1C                     ; '0'. Most significant digit.
        DEFB    $1C                     ; '0'.
        DEFB    $1C                     ; '0'.
        DEFB    $1C                     ; '0'. Least signifcant digit.

; --------------------------------------------
; Move the Invader Bomb and Laser Base Missile
; --------------------------------------------
; This routine takes between 4571 to 4898 T-states, although it is typically 4598 T-states.

; First move the invader bomb if there is one.

L4170:  LD      DE,$0020                ; 10
        LD      HL,($44F4)              ; 16    Fetch the invader bomb display file position.
        LD      A,H                     ; 4
        ADD     HL,DE                   ; 11    Address of the next row, where the bomb will move to.
        AND     A                       ; 4     Is there an active bomb?
        JR      NZ,L4181                ; 12/7  Jump ahead if there is.

; There is not an active invader bomb.

        LD      B,$04                   ; 7     Perform a delay to balance the other possible paths.

L417D:  DJNZ    L417D                   ; 13/8

        JR      L41AF                   ; 12    Jump ahead.

; There is an active invader bomb. Check whether it has reached row 22. If it has then its location
; within the display file will be at $47C0 or greater.

L4181:  CP      $47                     ; 7     Has the invader bomb potentially reached row 22?
        JR      NZ,L418F                ; 12/7  Jump ahead if not.

        LD      A,L                     ; 4     Has the invader bomb reached row 22?
        CP      $BF                     ; 7
        JR      C,L4192                 ; 12/7  Jump ahead if not.

; The invader bomb has reached row 22.

        PUSH    BC                      ; 11    Perform a delay to balance the alternate path to $41B1.
        POP     BC                      ; 10
        RET     C                       ; 11/5

        JR      L41B1                   ; 12    Jump ahead to continue.

; The invader bomb has not reached row 22.

L418F:  LD      A,(HL)                  ; 7     Perform a delay to balance the alternate path to $4192.
        LD      A,(HL)                  ; 7
        NOP                             ; 4

; The invader bomb has not reached row 22.

L4192:  LD      A,(HL)                  ; 7     Fetch the character at the new invader bomb position.
        AND     A                       ; 4     Has it hit a shield?
        JR      Z,L41AF                 ; 12/7  Jump ahead if so.

        RES     7,A                     ; 8     Make the character non-inverse.
        CP      $0D                     ; 7     Is it '$', i.e. the laser base?
        JR      NZ,L41AB                ; 12/7  Jump ahead if not.

; The invader bomb hit the laser base and so the game is over.

        POP     BC                      ; 10    Drop the return address.

; Update the last score store. The timing of these instructions is not important as the game will be restarted.

L419D:  LD      BC,$0004                ; 10    There are 4 digits used for the score.
        LD      HL,$47FB                ; 10    The address of the score digits.
        LD      DE,L416C                ; 10    The address of the last score store.
        LDIR                            ; 21/16 Update the last score.

        JP      L4000                   ; 10    Restart the game.

; The invader bomb has not hit the laser base or a shield so draw it.

L41AB:  LD      (HL),$94                ; 10    Draw the invader bomb '*'.

; Enters here from the main program loop for the current wave of invaders.

L41AD:  JR      L41B6                   ; 12    Jump ahead.

; The invader bomb has hit a shield.

L41AF:  LD      (HL),$80                ; 10    Remove a character from the shield.

L41B1:  LD      H,D                     ; 4     Perform a delay to balance the alternate path to $41AB.
        LD      L,D                     ; 4
        PUSH    BC                      ; 11
        POP     BC                      ; 10
        NOP                             ; 4

; The invader bomb has been moved. Six paths join here, each approximately taking the same amount of time to arrive:
;   Path 1 - There is not an active invader bomb (161 T-states).
;   Path 2 - The invader bomb has reached row 22 (160 T-states).
;   Path 3 - The invader bomb has hit a shield #1 (160 T-states).
;   Path 4 - The invader bomb has hit a shield #2 (160 T-states).
;   Path 5 - The invader bomb has not hit the laser base #1 (161 T-states).
;   Path 6 - The invader bomb has not hit the laser base #2 (161 T-states).

L41B6:  LD      ($44F4),HL              ; 16    Store the new invader bomb display file position.

; Now move the laser base missile if there is one.

        LD      HL,($44F6)              ; 16    Fetch the laser base missile display file position.
        LD      A,H                     ; 4
        AND     A                       ; 4     Is there an active missile?
        JR      NZ,L41C6                ; 12/7  Jump ahead if there is.

; There is not an active laser base missile.

        LD      B,$07                   ; 7     Perform a delay to balance the other possible paths.

L41C2:  DJNZ    L41C2                   ; 13/8

        JR      L41E5                   ; 12    Jump ahead.

; There is an active laser base missile. Check whether it has moved off the top of the screen. If it has then
; its location within the display file will be at $44FE or lower.

L41C6:  SBC     HL,DE                   ; 15    Address of the previous row, where the missile will move to.
        LD      A,H                     ; 4     Has the laser base missile reached row -1?
        CP      $44                     ; 7
        JR      NZ,L41D3                ; 12/7  Jump ahead if not.

; The laser base missile has moved off the top of the screen.

        ADD     IX,IX                   ; 15    Perform a delay to balance the alternate path to $41C2.

        LD      B,$02                   ; 7
        JR      L41C2                   ; 12

L41D3:  LD      A,(HL)                  ; 7     Fetch the character at the new missile location.
        CP      $94                     ; 7     Is it an invader bomb, i.e. an inverse '*'.
        JR      NZ,L41DE                ; 12/7  Jump ahead if not.

; The laser base missile has hit an invader bomb. Note that it is possible for both an invader bomb and the laser base missile
; to be moved past each other if they happened to be adjacent to each other. If there is a black cell between them then they will
; both be moved to the same location and hence will cancel each other out.

        LD      (IY+$1A),$00            ; 19    Clear the display file position of the invader bomb, i.e mark it as no active bomb.
        JR      L41E3                   ; 12    Jump ahead.

; The laser base missile has not hit an invader bomb.

L41DE:  ADD     IX,IX                   ; 15    Perform a delay.

        AND     A                       ; 4     Has the missile hit one of the the shields?
        JR      NZ,L41EA                ; 12/7  Jump ahead if not.

; The laser base missile has hit a shield or an invader bomb.

L41E3:  LD      (HL),$80                ; 10    Remove a character from the shield or erase the invader bomb.

L41E5:  LD      L,D                     ; 4     Perform a delay to balance the alternate path to $41F9.
        LD      H,D                     ; 4
        NOP                             ; 4

        JR      L41F9                   ; 12    Jump ahead.

; The laser base missile has not hit a shield or an invader bomb.

L41EA:  CP      $80                     ; 7     Is there a black space at the new laser base missile position?
        JR      Z,L41F7                 ; 12/7  Jump ahead if so.

; The laser base missile has hit an invader.

        CALL    L40DE                   ; 17    Erase the invader and increment the score, taking 365 T-states to 692 T-states.

        LD      (HL),$80                ; 10    Remove the laser base missile at its current position.
        LD      H,$00                   ; 7     This will result in the display file position of the laser base missile being marked as no active missile.
        JR      L41FE                   ; 12    Jump ahead.

; The laser base missile has not hit an invader or a shield or an invader bomb.

L41F7:  LD      (HL),$AE                ; 10    Draw the laser base missile, i.e. an inverse 'I'.

L41F9:  NOP                             ; 4     Perform a delay to balance the alternate path to $41F9.

        LD      B,$20                   ; 7

L41FC:  DJNZ    L41FC                   ; 13/8

; The laser base missile has been moved. Six paths join here, each approximately taking the same amount of time from $41B6 to arrive:
;   Path 1 - There is not an active laser base missile (598 T-states).
;   Path 2 - The laser base missile has move off screen (598 T-states).
;   Path 3 - The laser base missile has hit an invader bomb (598 T-states).
;   Path 4 - The laser base missile has hit a shield (598 T-states).
;   Path 5 - The laser base missile has moved to a blank position (598 T-states).
;   Path 6 - The laser base missile has hit an invader (207 + 365 to 692 T-states, i.e. 572 T-states to 899 T-states).

L41FE:  LD      ($44F6),HL              ; 16    Store the new laser base missile display file position.

; Perform a delay of 3823 T-states before returning. This ensures that the routine approximately matches
; the durations of the other routines called from the main loop.

        LD      A,(HL)                  ; 7
        LD      B,$25                   ; 7

L4204:  DJNZ    L4204                   ; 13/8

L4206:  DJNZ    L4206                   ; 13/8

        RET                             ; 10

; -------------------------------
; Move and Display the Laser Base
; -------------------------------
; This routine reads the keyboard and moves the laser base. It can move from column 2 to column 28 on row 21.
; It is moved to the right by key 8 or 3, and moved to the left by key 5 or 6, and key 0 or 1 will fire a missile.
; The routine takes between 4595 T-states to 4598 T-states, typically taking 4597 T-states.

L4209:  LD      L,(IY+$21)              ; 19    Fetch the low address of the laser base position within the display file.
        LD      H,$47                   ; 7     The high display file address of the laser base.
        LD      (HL),$80                ; 10    Erase the laser base.

; Upper byte selects keyboard row to read.
;
;          Bit0  Bit1  Bit2  Bit3  Bit4    Bit4  Bit3  Bit2  Bit1  Bit0
;          ----  ----  ----  ----  ----    ----  ----  ----  ----  ----
; $F7FE =    1     2     3     4     5       6     7     8     9     0   = $EFFE
; $FBFE =    Q     W     E     R     T       Y     U     I     O     P   = $DFFE
; $FDFE =    A     S     D     F     G       H     J     K     L    N/L  = $BFFE
; $FEFE =  SHIFT   Z     X     C     V       B     N     M     .   SPACE = $7FFE

        LD      B,$E7                   ; 7     Select both halves of the top keyboard row (1-5 and 6-0).
        LD      C,$FE                   ; 7     The keyboard I/O port.
        LD      A,L                     ; 4     Fetch the low byte of the laser base display file address.
        IN      D,(C)                   ; 12    Read the keyboard.

        BIT     2,D                     ; 8     Was key 8 (or 3) pressed?
        JR      NZ,L4222                ; 12/7  Jump ahead if not.

; Key 8 (or 3) was pressed so attempt to move to the right.

        CP      $BC                     ; 7     Is the laser base at column 28?
        JR      Z,L4220                 ; 12/7  Jump ahead if so.

; The laser base is not yet at the right hand edge of the screen so it can be moved.

        INC     L                       ; 4     Increment the laser base display file address, i.e. one cell to the right.

L4220:  JR      L4225                   ; 12    Jump ahead to continue.

; Key 8 (or 3) was not pressed. Delay to balance other paths through this routine and then test for other key presses.

L4222:  INC     (HL)                    ; 11    Perform a delay to balance the alternate paths to $4225.
        DEC     (HL)                    ; 11
        NOP                             ; 4

L4225:  BIT     4,D                     ; 8     Was key 5 (or 6) pressed?
        JR      NZ,L4230                ; 12/7  Jump ahead if not.

; Key 5 (or 6) was pressed so attempt to move to the left.

        CP      $A2                     ; 7     Is the laser base at column 2?
        JR      Z,L422E                 ; 12/7  Jump ahead if so.

; The laser base is not yet at the left hand edge of the screen so it can be moved.

        DEC     L                       ; 4     Decrement the laser base display file address, i.e. one cell to the left.

L422E:  JR      L4233                   ; 12    Jump ahead to continue.

; Key 5 (or 6) was not pressed. Delay to balance other paths through this routine and then test for other key presses.

L4230:  INC     (HL)                    ; 11    Perform a delay to balance the alternate paths to $4233.
        DEC     (HL)                    ; 11
        NOP                             ; 4

L4233:  LD      (IY+$21),L              ; 19    Store the new laser base display file address.
        LD      (HL),$8D                ; 10    Display the laser base as an inverse '$' to indicate that there is an active missile.

        LD      A,($44F7)               ; 13    Is there an active laser base missile?
        AND     A                       ; 4
        JR      Z,L4242                 ; 12/7  Jump if there is not.

; There is an active laser base missile, so there is no need to test for the fire key since it is only possible
; to fire one missile at a time.

        INC     HL                      ; 6     Perform a delay to balance the alternate path to $4246.
        DEC     HL                      ; 6
        JR      L4246                   ; 12    Jump ahead to continue.

; There is not an active laser base missile.

L4242:  LD      (HL),$0D                ; 10    Display the laser base as a '$' to indicate that there is not an active missile.
        BIT     0,D                     ; 8     Was key 0 (or 1) pressed?

L4246:  JR      NZ,L4251                ; 12/7  Jump ahead if not.

; Key 0 (or 1) was pressed so fire a missile.

        RES     5,L                     ; 8     Generate the display file address of the character above the laser base.
        LD      (HL),$AE                ; 10    Display a missile, i.e. an inverse 'I'.
        LD      ($44F6),HL              ; 16    Store the address within the display file of the missile.

        JR      L4256                   ; 12    Jump ahead to continue.

; Key 0 (or 1) was not pressed. Delay to balance other paths through this routine.

L4251:  INC     (HL)                    ; 11    Perform a delay to balance the alternate paths to $4256.
        DEC     (HL)                    ; 11
        ADD     IX,IX                   ; 15
        NOP                             ; 4

; Perform a delay to ensure this routine takes a certain amount of time to execute.

L4256:  LD      B,$4B                   ; 7     Perform a delay of 4310 T-states.

L4258:  DJNZ    L4258                   ; 13/8

L425A:  DJNZ    L425A                   ; 13/8

        RET                             ; 10

; -------------------------------------
; Check Whether to Drop an Invader Bomb
; -------------------------------------
; This routine searches the invader area from row 13 up to row 0 at the column where the laser base is.
; It searches for an invader that is directly above the laser base and sets a flag when one is found. Since
; all rows are checked, this flag ensures that the lowest invader found is the one that drops the bomb.
; Note that the first instructions of the routine can result is writes to the first page of the ROM area when
; there is no active invader bomb or laser base missile.
; The routine takes between 4570 T-states and 4598 T-states, typically taking 4598 T-states.

L425D:  LD      HL,($44F6)              ; 16    Fetch the laser base missile display file address.
        LD      (HL),$80                ; 10    Erase the missile (if there is no active missile then this writes to the first page of ROM).

        LD      HL,($44F4)              ; 16    Fetch the invader bomb display file address.
        LD      (HL),$80                ; 10    Erase the bomb (if there is no active bomb then this writes to the first page of ROM).

        LD      DE,$0020                ; 10    An offset to the previous row.

        LD      L,(IY+$21)              ; 19    Fetch the low address of the laser base display file address.
        LD      H,$46                   ; 7     HL points to the column above the laser base but on row 13 (the active last row of the invader area).

        LD      (IY+$20),$00            ; 19    Signal that an invader has not yet been selected to drop a bomb.

; Enter a loop to test each row of the invader area to see if there is an invader directly overhead the laser base.

L4273:  LD      A,(HL)                  ; 7     Fetch the character within the invader area above the laser base.
        SUB     $1C                     ; 7     Convert it to a numerical value, i.e. subtract '0'. Will yield 0 to 9 if an invader is directly overhead the laser base.
        BIT     7,A                     ; 8     Could it be a number character?
        JR      NZ,L4295                ; 12/7  Jump ahead if not.

; There might be an invader directly overhead the laser base.

        SUB     $0A                     ; 7     If an invader is directly overhead the laser base then this test will result in a negative value.
        BIT     7,A                     ; 8     Was it a number character, i.e. an invader directly above the laser base?
        JR      Z,L4297                 ; 12/7  Jump ahead if not.

; An invader is directly above the laser base.

        LD      A,($44F5)               ; 13    Fetch the high byte of the invader bomb display file address.
        AND     A                       ; 4     Is there an active bomb?
        JR      NZ,L429A                ; 12/7  Jump ahead if so.

; There is no active invader bomb.

        LD      A,($44FB)               ; 13    Has an invader already been selected to drop a bomb?
        AND     A                       ; 4
        JR      NZ,L429D                ; 12/7  Jump ahead if so.

; No other invader has yet been selected to drop a bomb so choose this one.

        ADD     HL,DE                   ; 11    The address of the cell beneath the invader.
        LD      ($44F4),HL              ; 16    Store the new invader bomb display file address.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point back to the invader.
        JR      L42AC                   ; 12    Jump ahead to continue.

; An invader is not directly overhead the laser base.

L4295:  INC     (HL)                    ; 11    Perform a delay to balance the alternate path to $4297.
        DEC     (HL)                    ; 11

; Join here if the invader being considered is not directly overhead the laser base.

L4297:  PUSH    BC                      ; 11    Perform a delay to balance the alternate paths to $429A.
        POP     IX                      ; 14

; Join here if there is already an active invader bomb.

L429A:  PUSH    BC                      ; 11    Perform a delay to balance the alternate paths to $429C.
        POP     IX                      ; 14

; Join here if an invader has already been selected to drop a bomb.

L429D:  LD      A,(HL)                  ; 7     Does this location contain an invader?
        CP      $80                     ; 7
        JR      NZ,L42A6                ; 12/7  Jump ahead if it does.

; There is no invader at the current location.

        JR      L42A4                   ; 12    Jump to the following instruction thereby performing a delay.

L42A4:  JR      L42AA                   ; 12    Jump ahead to continue.

; The current location contains an invader directly overhead the laser base and there is no active bomb.

L42A6:  LD      (IY+$20),$FF            ; 19    Signal that an invader has been selected to drop a bomb.


L42AA:  NOP                             ; 4     Perform a delay.
        NOP                             ; 4

L42AC:  AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point to the previous row within the invader area, but still directly above the laser base.

        LD      A,H                     ; 4
        CP      $44                     ; 7     Have all of the rows within the invader area been considered?
        JR      NZ,L4273                ; 12/7  Jump back if not to test the previous row.

; All rows within the invader area have been tested to see whether an invader can drop a bomb.
; Now test whether the invaders have reached the row immediately above the shields and hence the game is lost.

        LD      HL,$46BF                ; 10    Row 13, column 31. This will point at the newline character terminating row 13.
        LD      B,$1F                   ; 7     31 columns.

L42B9:  INC     HL                      ; 6     Next lcoation.
        LD      A,(HL)                  ; 7
        CP      $80                     ; 7     Does the location contain a black space?
        JP      NZ,L419D                ; 10    Jump if a non-black space character was found, i.e. the invaders have reached the shields.
                                        ;       [Ideally this jump should have been to $419C to ensure that the call return address is discarded.
                                        ;       However the stack pointer gets reset and so there is not a stack leak.]

        DJNZ    L42B9                   ; 13/8  Repeat for all columns in row 13.

; The invaders have not reached the level of the shields.

        LD      B,$19                   ; 7     Perform a delay.

L42C4:  DJNZ    L42C4                   ; 13/8

        RET                             ; 10

; ======================================================================================================================================================

; ===========
; FLOW CHARTS
; ===========

; -------------
; Move Invaders
; -------------
; $4110. The routine takes between 4589 T-states and 4607 T-states, although it is typically 4597 T-states.
;
;
;                               $4110                                MOVE ROW OF INVADERS
;                                 |
;                              15 |
;                                 v         12
;                                 +-----------------------+
;                               7 |   Move Invaders Left  |
;                                 v                       v
;                              22 |                       | 17
;                                 v                       |
;                           $41FF +<----------------------+
;                                 |
;                               7 |
;                                 v
;         +---------------------->+ $4120
;         |                       |
;         |                     7 |
;         |                       v         12
;         |                       +---------------------->+ $4129
;         |                     7 |   Not Diagonal Chr$   |
;         |                       v                       | 11
;         |                       |                       v         12
;         |              |        |                       +-----------------------+
;         |              |        |                       |   Not Diagonal Chr$   |
;         |       30     |     26 |                     7 |                       |
;         |   Iterations |        |                       v                       |
;         |              |        |                     7 |                       |
;         |              |        v                       v                       |
;         |              |  $4130 +<----------------------+<----------------------+
;         |              v        |
;         |                     8 |
;         |                       v         12
;         |                       +-----------------------+
;         |                     7 |   Move Invaders Down  |
;         |                       v                       v
;         |                    57 |                       | 52
;         |                       v                       |
;         |                 $4144 +<----------------------+
;         |                       |
;         |                       | 29
;         |  End Row Not Reached  v
;         +-----------------------+
;                   12            | 7
;                                 v
;                                 | 74
;                                 v         12
;                                 +-----------------------+
;                               7 |   More Area To Shift  |
;                                 v                       |
;                              18 |                       |
;                   12            v                       v
;         +-----------------------+                       | 7
;         |  Not At Screen Edge   | 7                     v
;         |                       v                       |
;         |                       | 7                     |
;         |                 $4162 v                       | 3*13-5
;         +---------------------->+                       |
;                                 | 16                    |
;                                 v                       |
;                           $4169 +<----------------------+
;                                 |
;                              18 |
;                                 v
;                              RETURN

; --------------------------------------------
; Move the Invader Bomb and Laser Base Missile
; --------------------------------------------
; $4170. The routine takes between 4571 to 4899 T-states, although it is typically 4598 T-states.
;
;
;                               $4170                                   MOVE INVADER BOMB
;                                 |
;                              49 |
;                                 v          7
;                                 +-----------------------------------------------+
;                Active Bomb   12 |                                               |
;                                 v                                               v
;                               7 |                                               |
;                                 v         12                                    | 7
;                                 +-----------------------+                       |
;                               7 |    Above Row 22       |                       v
;                                 v                       v $418F                 |
;                                 |                       |                       | (4*13)-5
;                              11 |                       | 18                    |
;                                 v         12            v                       v
;                                 +---------------------->+ $4192                 |
;                                 |    Above Row 22       |                       | 12
;                               7 |                       | 11                    |
;                                 v                       v         12            v
;                                 |                       +---------------------->+ $41AF
;                              43 |                     7 |    Hit Shield         |
;                                 |                       v                       | 10
;                                 v                       |                       |
;                           $41B1 +<----------------------=-----------------------+
;                                 |                       |
;                                 |                    15 |
;                                 |                       v          7
;                                 |                       +-----------------------+
;                                 |                       |    Hit Laser Base     |
;                              33 |                    12 |                       | 50
;                                 |                       v $41AB                 v
;                                 |                       |                       |
;                                 |                    10 |                       | (21*4)-5
;                                 |                       v $41AD                 |
;                                 |                       |                       |
;                                 |                    12 |                       v
;                                 v                       |                  Restart Game
;                           $41B6 +<----------------------+                     $4000
;                                 |
;                              16 |
;                                 v
;                              28 |                               MOVE LASER BASE MISSILE
;                                 v          7
;                                 +-----------------------------------------------+
;             Active Missile   12 |                                               |
;                                 v $41C6                                         |
;                              26 |                                               v
;                                 v          7                                    |
;                                 +-----------------------+                       | 7
;                                 |    Off Screen         |                       |
;                              12 |                       |                       |
;                                 v $41D3                 v                       v
;                              14 |                       |                       |
;                   12            v                       | 34                    |
;         +-----------------------+                       |                       |
;         |                       |                       v $41C2                 |
;   $41DE v                     7 |  Hit Invader Bomb     |                       |
;         |                       v                       |                       |
;      19 |                    31 |                       |                       |
;         v          7            v $41E3                 | (2*13)-5              | (7*13)-5
;         +---------------------->+                       |                       |
;         |    Hit Shields        |                       |                       |
;      12 |                    10 |                       |                       |
;         |                       v         12            v                       |
;         v $41EA           $41E5 +<----------------------+<----------------------+
;         |                       |
;       7 |                       | 28
;         v    12         10      v
;         +---------->----------->+ $41F9
;       7 |        $41F7          |
;         v                       | 11
;      17 |                       |
;         v  Hit Invader          v
;         |                       |
;     CALL $40DE (365->692)       |
;         |                       | (32*13)-5
;         v                       |
;      29 |                       v
;         +---------------------->+ $41FE
;                                 |
;                                 | 30
;                                 v
;                                 | (37*13)-5
;                                 v
;                                 | (256*13)-5
;                                 v
;                                 | 10
;                                 v
;                              RETURN

; -------------------------------------
; Laser Base Missile Has Hit An Invader
; -------------------------------------
; $40DE. The routine typically takes 365 T-states, but can be up to 692 T-states.
;
;
;                               $40DE                                 MISSILE HIT INVADER
;                                 |
;                              15 |
;                                 v
;                           $40E2 +<----------------------------------------------------------------------+
;                                 |                                                                       |
;                              14 |                                                                       |
;                                 v         10                                                            |
;                                 +-----------------------------------------------+                       |
;                              10 |   Not Central Chr$                            |                       |
;                                 v                                               |                       |
;                               7 |                                               |                       |
;                                 v         10                                    v                       |
;                                 +---------------------------------------------->+                       |
;                              10 |   Not Central Chr$                            |                       |
;                                 v                                               |                       |
;         Invader Value Found     |                                               |                       |
;                              14 |                                               |                       |
;                                 v                                               |                       |
;         +---------------------->+ $40F1                                         |                       |
;         |                       |                                               |                       |
;         |                    18 |                                               |                       |
;         |        4      |       v         10                                    |                  ^    |
;         |    Potential  |       +-----------------------+                       |                  |    |
;         |    Iterations |    10 |   No Digit Overflow   |                       |           4      |    |
;         |               v       v                       |                       |       Iterations |    |
;         |                    36 |                       |                       |                  |    |
;         |         12            v                       |                       |                  |    |
;         +-----------------------+                       |                       |                       |
;         Increment Next Digit    | 7                     |                       |                       |
;                                 v                       |                       |                       |
;                                 | 4                     |                       |                       |
;                                 v                       |                       |                       |
;                           $4104 +<----------------------+                       |                       |
;                                 |                                               |                       |
;                                 | 7                                             |                       |
;                                 v                                               |                       |
;                           $4105 +<----------------------------------------------+                       |
;                                 |                                                                       |
;                                 | 14                                                                    |
;                                 v         13                                                            |
;                                 +-----------------------------------------------------------------------+
;                               8 |   Check Next Invader Chr$
;                                 v
;                              23 |
;                                 v         12
;                                 +-----------------------+
;                                 |                       |
;                               7 |                       v $40C9
;                                 |                     4 |
;                                 v                       v
;                                 |                    24 |
;                              10 |                       v
;                                 |                    10 |
;                                 v                       v
;                              RETURN             New Invader Wave
;                                                      $4062

; -------------------------------
; Move and Display the Laser Base
; -------------------------------
; $4209. The routine takes between 4595 T-states to 4598 T-states, although it is typically 4597 T-states.
;
;
;                               $4209                                     MOVE LASER BASE
;                                 |
;                              74 |
;                                 v         12
;                                 +-----------------------------------------------+
;                               7 |                                               |
;               Key 8 Pressed     v                                               |
;                               7 |                                               |
;                                 v         12                                    v
;                                 +-----------------------+                       |
;                               7 |    Base At Column 28  |                       |
;                                 v                       |                       |
;                               4 |                       |                       | 26
;                                 v                       |                       |
;                           $4220 +<----------------------+                       |
;                                 |                                               |
;                              12 |                                               |
;                                 v                                               |
;                           $4225 +<----------------------------------------------+
;                                 |
;                               8 |
;                                 v         12
;                                 +-----------------------------------------------+
;                               7 |                                               |
;               Key 5 Pressed     v                                               |
;                               7 |                                               |
;                                 v         12                                    v
;                                 +-----------------------+                       |
;                               7 |    Base At Column 2   |                       |
;                                 v                       |                       |
;                               4 |                       |                       | 26
;                                 v                       |                       |
;                           $422E +<----------------------+                       |
;                                 |                                               |
;                              12 |                                               |
;                                 v                                               |
;                           $4233 +<----------------------------------------------+
;                                 |
;                              46 |
;                                 v         12
;                                 +-----------------------+
;                               7 |    No Active Missile  |
;                                 v                       v
;                              24 |                       | 18
;                   12            v                       |
;         +-----------------------+<----------------------+
;         |                 $4246 |
;         |                       | 7
;         v                       v    Key 0 Pressed
;         |                       |
;      41 |                       | 46
;         |                       v
;         +---------------------->+ $4256
;                                 |
;                                 | 7
;                                 v
;                                 | (75*13)-5
;                                 v
;                                 | (256*13)-5
;                                 v
;                                 | 10
;                                 v
;                              RETURN

; -------------------------------------
; Check Whether to Drop an Invader Bomb
; -------------------------------------
; $425D. The routine takes between 4570 T-states and 4598 T-states, although it is typically 4598 T-states.
;
;
;                               $425D                      CHECK WHETHER TO DROP NEW BOMB
;                                 |
;                                 | 107
;                                 v
;         +---------------------->+ $4273
;         |                       |
;         |                       | 22
;         |                       v         12
;         |                       +-----------------------+ $4295
;         |                     7 |   Not Number Chr$     |
;         |                       v                       | 22
;         |                    15 |                       |
;         |                       v         12            v
;         |                       +---------------------->+ $4297
;         |                     7 |   Not Number Chr$     |
;         |              |        v                       | 25
;         |              |     17 |                       |
;         |       14     |        v         12            v
;         |   Iterations |        +---------------------->+ $429A
;         |              |      7 |   Active Bomb Exists  |
;         |              |        v                       | 25
;         |              v     17 |                       |
;         |                       v         12            v
;         |                       +---------------------->+ $429D
;         |                     7 |   Bomb Already        |
;         |                       v   Selected            | 14
;         |                       |                       v         12
;         |                       |                       +-----------------------+ $42A6
;         |       Select Invader  |                     7 |   Found Invader       |
;         |       To Drop Bomb    |                       v                       |
;         |                       |                    12 |                       | 19
;         |                    58 |                       v                       |
;         |                       |                    12 |                       |
;         |                       |                       v                       |
;         |                       |                 $42AA +<----------------------+
;         |                       |                       |
;         |                       v                       | 8
;         |                 $42AC +<----------------------+
;         |                       |
;         |      More Rows     30 |
;         |      To Check         v
;         +-----------------------+
;                   12            | 7
;                                 v
;                                 | 17                 CHECK WHETHER INVADERS HAVE LANDED
;                                 v
;         +---------------------->+ $42B9
;         |                       |
;         |              |        | 20
;         |       31     |        v            10
;      13 |   Iterations |        +-----------------------> $419D (Invaders Landed)
;         |              v        |
;         |                       | 10
;         | Check Next Column     v
;         +-----------------------+
;                                 | 8
;                                 v
;                                 | 7
;                                 v
;                                 | (25*13)-5
;                                 v
;                                 | 10
;                                 v
;                              RETURN

; ======================================================================================================================================================

        END
